#include "boolean.h"
#include "compute_schedule.h"
#include "region.h"
#include "pgme.h"
#include "schedule.h"
#include "mbp2bdecomp.h"
#include "common.h"
#include "pvm_start.h"
#include <stdio.h>
#include <pvm3.h>

#define INIT_MSG 2000
#define MSGSIZE 500

int main(int nb_arg,char **arg) 
{
  int ii;
  char *pgme_name = "Pgme1";
  char *other_pgme_name = "Pgme2";
  pgme_s *my_pgme;
  pgme_s *other_pgme;
  int numNode_other;
  int me = -1;
  int numNode = -1;

  DARRAY *BP_da;

  double *BP_array;

  int i, j, k;
  int size[BP_Ndim]; 
  
  region_s *r;
  setOfRegion_s *BP_Set;
  
  int left[BP_Ndim];
  int right[BP_Ndim];
  int stride[BP_Ndim];
  int numproc;
  int da_size[BP_Ndim];
  
  boolean compute_remapping = true;
  boolean new_iteration = true;
  
  sched_s *sched;

  decomp_Birreg* irr_dd;
  
  int IterationCounter = 0;
  
  int SetRegionOverAllSize;
  int OverAllSize;
  
  int gb1, gb2;
  
  int LZ_count;

  int BP_DIM1 = 1024, BP_DIM2 = 1024; 
  
  if (nb_arg < 3) {
    printf("Usage main1 #replicate #friends\n");
    exit(1);
  }
  
  numNode_other = atoi(arg[2]);
  
  if (numNode_other <= 0 || atoi(arg[1]) <= 0) {
    printf("Wrong value for one of the params\n");
    exit(1);
  }
  
  numNode = start_master_slave(arg);
  if (numNode < 2)
    exit(0);
  
  /*
   * Initialise 
   *    the pvm library
   *    the chaos library
   */
  
  parti_pvm_setup(numNode, 100, pgme_name);
  PARTI_setup();

  printf("Check 0\n"); 

  /*
   * Initialise the pgme structure 
   * who am i, who are the other nodes ...
   */  
  my_pgme = InitPgme("Truc1", numNode);
  other_pgme = WaitPgme("Truc2", numNode_other);

  printf("Check 1\n");

  /*
   * pvm inquiries
   */
  numproc = NumNodePgme(my_pgme);
  me = MyPosMyPgme();

  /*
   * Read the meshes file 
   */
  if (me == 0)
    {
      pvm_initsend(PvmDataDefault);
      pvm_pkint(&BP_DIM1, 1, 1);
      pvm_pkint(&BP_DIM2, 1, 1);
      pvm_bcast("Truc2", INIT_MSG);
    }
  
  init_timers();
  
  Sync2Pgme(my_pgme, other_pgme);

  /* Create a DAD for Block Parti */
  BP_da = BP_create_darray(BP_DIM1, BP_DIM2);  
  
  /* Allocate and initialize the BP_array */
  laSizes(BP_da, da_size);
  
  BP_array = (double*)malloc(sizeof(double)*da_size[0]*da_size[1]);
  
  /* Change Regular MB to Irregular and Initialize(need to interpret Data Descriptor */
  irr_dd = DARRAY2BirregTree(BP_da);

  Initialize_distributed_array(irr_dd,BP_array,BP_DIM1,BP_DIM2);

  /* create the SetOfRegion for BP */
  BP_Set = Alloc_setOfRegion();
  LZ_count = 0;

  /********************** 16 regions (1024x1024)***********************/
  
  /* half data */
  
  /* region 1 */
  
  left[0] = 0;    left[1] = 100;
  right[0] = 299; right[1] = 199;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);

  /* region 2 */
  
  left[0] = 0;    left[1] = 200;
  right[0] = 149; right[1] = 499;
  stride[0] = 1;  stride[1] = 1;
  
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 3 */
  left[0] = 150;  left[1] = 500;
  right[0] = 299; right[1] = 899;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 4 */
  
  left[0] = 0;    left[1] = 900;
  right[0] = 299; right[1] = 961;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 5 */

  left[0] = 400;  left[1] = 0;
  right[0] = 499; right[1] = 199;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 6 */
  
  left[0] = 300;  left[1] = 200;
  right[0] = 499; right[1] = 349;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 7 */
  
  left[0] = 400;  left[1] = 500;
  right[0] = 499; right[1] = 899;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 8 */
  
  left[0] = 300;  left[1] = 962;
  right[0] = 499; right[1] = 1023;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);

  /* region 9 */
  
  left[0] = 500;  left[1] = 0;
  right[0] = 549; right[1] = 199;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
    
  /* region 10 */

  left[0] = 550;  left[1] = 200;
  right[0] = 599; right[1] = 499;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
    
  /* region 11 */

  left[0] = 500;  left[1] = 500;
  right[0] = 599; right[1] = 699;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
 
  /* region 12 */
  
  left[0] = 500;  left[1] = 962;
  right[0] = 599; right[1] = 1023;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
   
  /* region 13 */

  left[0] = 600;   left[1] = 0;
  right[0] = 1023; right[1] = 99;
  stride[0] = 1;   stride[1] = 1;
  
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 14 */
  
  left[0] = 600;   left[1] = 350;
  right[0] = 1023; right[1] = 499;
  stride[0] = 1;   stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 15 */

  left[0] = 600;   left[1] = 700;
  right[0] = 1023; right[1] = 899;
  stride[0] = 1;   stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);
  
  /* region 16 */
  
  left[0] = 600;  left[1] = 900;
  right[0] = 811; right[1] = 1023;
  stride[0] = 1;  stride[1] = 1;
 
  r = Alloc_R_HPF(2, left, right, stride, R_replicated, Mem_Copy);
  Add_Region_setOfRegion(r, BP_Set);

  /****************************************/

  Init_PVM_routes(my_pgme, other_pgme);

  Sync2Pgme(my_pgme, other_pgme);
  usleep(9000);
  start_timer_1(1);

  OverAllSize = SizeSetOfRegion(BP_Set);
  
  /******************** New Compute Schedules ********************/ 
  
  /* 1. All to All, Both Sides */
  
  sched = ComputeSchedule_Two_Compacts(SendAllToAll, irr_dd, BP_Set, my_pgme, other_pgme);
  
  /* 2. One to One, Both sides */ 
  
  /* sched = ComputeSchedule_Two_Compacts(SendOneToOne, irr_dd, BP_Set, my_pgme, other_pgme); */  
  
  /* 3. All to All , OneSideOnMe */
  
  /* sched = ComputeScheduleOnMe_Two_Compacts(SendAllToAll, irr_dd, BP_Set, my_pgme, other_pgme); */
  
  /* 4. One to One, OneSideOnMe */
  
  /* sched = ComputeScheduleOnMe_Two_Compacts(SendOneToOne, irr_dd, BP_Set, my_pgme, other_pgme); */
  
  /* 5. All to All, OneSideOnOther */
  
  /* sched = ComputeScheduleOnOther_Two_Compacts(SendAllToAll, irr_dd, BP_Set, my_pgme, other_pgme); */
  
  /* 6. One to One, OneSideOnOther */
  
  /* sched = ComputeScheduleOnOther_Two_Compacts(SendOneToOne, irr_dd, BP_Set, my_pgme, other_pgme); */    
    
  stop_timer(1);
  
  Sync2Pgme(my_pgme, other_pgme); 
   
  /********** Move Data *****************/
  
  start_timer_1(2);
  
  dDataMoveSend(other_pgme, sched, BP_array, 1001);

  stop_timer(2);
  
  Sync2Pgme(my_pgme, other_pgme);
    
  /***************************************/
  
  printf("\n I am the process %d in main1\n",me); 
  
  {
    FILE *OutFile;
    
    FILE *fp_max_scheds;
    FILE *fp_avg_scheds;
    
    FILE *fp_max_move;
    FILE *fp_avg_move;
    
    int min1, my1, max1, avg1;
    int min2, my2, max2, avg2;
    
    if (me == 0) 
      {
	OutFile = fopen("/tmp/result_mb2mb.1", "a");
	if (OutFile == NULL)
	  {
	    printf("\n /tmp/result_mb2mb.1 not open \n");
	    OutFile = stdout;
	  }
	
	fp_max_scheds = fopen("/tmp/max_sched.1", "a");
	if (fp_max_scheds == NULL)
	  {
	    printf("\n /tmp/max_sched.1 not open \n");
	    fp_max_scheds = stdout;
	  }
	
	fp_avg_scheds = fopen("/tmp/avg_sched.1", "a");
	if (fp_avg_scheds == NULL)
	  {
	    printf("\n /tmp/avg_sched.1 not open \n");
	    fp_avg_scheds = stdout;
	  }
	
	fp_max_move = fopen("/tmp/max_move.1", "a");
	if (fp_max_move == NULL)
	  {
	    printf("\n /tmp/max_move.1 not open \n");
	    fp_max_move = stdout;
	  }
	
	fp_avg_move = fopen("/tmp/avg_move.1", "a");
	if (fp_avg_move == NULL)
	  {
	    printf("\n /tmp/avg_move.1 not open \n");
	    fp_avg_move = stdout;
	  }
      }
    else
      OutFile = stdout;
    
    fprintf(OutFile, "\n-------------------------------------");
    fprintf(OutFile, "\n BP_Dim %i %i ", BP_DIM1, BP_DIM2);
    fprintf(OutFile, "\n me %i ", me);
    fprintf(OutFile, "\n numproc %i ", NumNodePgme(my_pgme));
    fprintf(OutFile, "\n numproc_other %i ", NumNodePgme(other_pgme));
    fprintf(OutFile, "\n Timers ");

    min1 = return_timer_min(1);
    my1 = return_timer(1);
    max1 = return_timer_max(1);
    avg1 = return_timer_avg(1);

    min2 = return_timer_min(2);
    my2 = return_timer(2);
    max2 = return_timer_max(2);
    avg2 = return_timer_avg(2);
    
    fprintf(OutFile, "\n                        =      min     mine      max      avg",
	    min1, my1, max1, avg1);
    fprintf(OutFile, "\n Compute Sched MB -> MB = %8i %8i %8i %8i",
	    min1, my1, max1, avg1);
    fprintf(OutFile, "\n remap                  = %8i %8i %8i %8i",
	    min2, my2, max2, avg2);
    
    fprintf(OutFile, "\n-------------------------------------");
    
    if (me == 0)
      {
        fprintf(fp_max_scheds, "\n%d", max1);
        fprintf(fp_max_move, "\n%d", max2);
	
        fprintf(fp_avg_scheds, "\n%d", avg1);
        fprintf(fp_avg_move, "\n%d", avg2);
      }

    fclose(OutFile);

    if (me == 0)
      {
        fclose(fp_max_scheds);
        fclose(fp_avg_scheds);
        fclose(fp_max_move);
        fclose(fp_avg_move);
      }
  }
  
  Sync2Pgme(my_pgme, other_pgme);
  
  if (me == 0)
    pvm_exit();
  
  printf("\n Finished normally in main1(me=%d)\n", me);
}
